@@ -1,3 +1,7 @@
+# SPDX-FileCopyrightText: 2021-2022 Constantine Evans <const@costi.eu>
+#
+# SPDX-License-Identifier: AGPL-3.0-only
+
 from __future__ import annotations
 
 from abc import ABCMeta, abstractmethod
@@ -26,6 +30,11 @@ class Normalizer(metaclass=ABCMeta):
     def normalize(self, data: pd.DataFrame) -> pd.DataFrame:
         ...
 
+    @property
+    @abstractmethod
+    def ylabel(self) -> str:  # pragma: no cover
+        ...
+
 
 class NormRaw(Normalizer):
     """
@@ -39,6 +48,8 @@ class NormRaw(Normalizer):
     def normalize(self, data: pd.DataFrame) -> pd.DataFrame:
         return data
 
+    ylabel = "fluorescence"
+
 
 @dataclass(init=False)
 class NormToMeanPerWell(Normalizer):
@@ -111,6 +122,8 @@ class NormToMeanPerWell(Normalizer):
 
         return normdata
 
+    ylabel = "norm. fluorescence"
+
 
 @dataclass
 class NormToMaxPerWell(Normalizer):
@@ -182,3 +195,5 @@ class NormToMaxPerWell(Normalizer):
         normdata.loc[:, (slice(None), "fl")] /= means
 
         return normdata
+
+    ylabel = "norm fluorescence"
